@@ -373,6 +373,7 @@ Ring::freeze_row(row_t position,
 	if (!row->attr.soft_wrapped)
 		g_string_append_c (buffer, '\n');
 	record.soft_wrapped = row->attr.soft_wrapped;
+        record.bidi_flags = row->attr.bidi_flags;
 
 	_vte_stream_append(m_text_stream, buffer->str, buffer->len);
 	append_row_record(&record, position);
@@ -435,6 +436,7 @@ Ring::thaw_row(row_t position,
                 g_string_truncate (buffer, buffer->len - 1);
 	else
 		row->attr.soft_wrapped = TRUE;
+        row->attr.bidi_flags = records[0].bidi_flags;
 
 	p = buffer->str;
 	end = p + buffer->len;
@@ -607,6 +609,27 @@ Ring::index(row_t position)
 	return &m_cached_row;
 }
 
+bool
+Ring::is_soft_wrapped(row_t position)
+{
+        const VteRowData *row;
+        RowRecord record;
+
+        if (G_UNLIKELY (position < m_start || position >= m_end))
+                return false;
+
+        if (G_LIKELY (position >= m_writable)) {
+                row = get_writable_index(position);
+                return row->attr.soft_wrapped;
+        }
+
+        /* The row is scrolled out to the stream. Save work by not reading the actual row.
+         * The requested information is readily available in row_stream, too. */
+        if (G_UNLIKELY (!read_row_record(&record, position)))
+                return false;
+        return record.soft_wrapped;
+}
+
 /*
  * Returns the hyperlink idx at the given position.
  *
@@ -726,7 +749,8 @@ Ring::discard_one_row()
 void
 Ring::maybe_freeze_one_row()
 {
-        if (G_LIKELY(m_mask >= m_visible_rows &&
+        /* See the comment about m_visible_rows + 1 at ensure_writable_room(). */
+        if (G_LIKELY(m_mask >= m_visible_rows + 1 &&
                      m_writable + m_mask + 1 == m_end))
 		freeze_one_row();
 	else
@@ -747,7 +771,11 @@ Ring::ensure_writable_room()
 	row_t new_mask, old_mask, i, end;
 	VteRowData* old_array, *new_array;;
 
-        if (G_LIKELY(m_mask >= m_visible_rows &&
+        /* Keep at least m_visible_rows + 1 rows in the ring.
+         * The BiDi spec requires that the just scrolled out row
+         * is still alterable (can be switched to hard line ending).
+         * It's nice anyway to make that hard wrapped upon a clear. */
+        if (G_LIKELY(m_mask >= m_visible_rows + 1 &&
                      m_writable + m_mask + 1 > m_end))
 		return;
 
@@ -756,7 +784,7 @@ Ring::ensure_writable_room()
 
 	do {
 		m_mask = (m_mask << 1) + 1;
-        } while (m_mask < m_visible_rows || m_writable + m_mask + 1 <= m_end);
+        } while (m_mask < m_visible_rows + 1 || m_writable + m_mask + 1 <= m_end);
 
 	_vte_debug_print(VTE_DEBUG_RING, "Enlarging writable array from %lu to %lu\n", old_mask, m_mask);
 
@@ -844,7 +872,7 @@ Ring::shrink(row_t max_len)
  * Return: the newly added row.
  */
 VteRowData*
-Ring::insert(row_t position)
+Ring::insert(row_t position, guint8 bidi_flags)
 {
 	row_t i;
 	VteRowData* row, tmp;
@@ -867,6 +895,7 @@ Ring::insert(row_t position)
 
 	row = get_writable_index(position);
 	_vte_row_data_clear (row);
+        row->attr.bidi_flags = bidi_flags;
 	m_end++;
 
 	maybe_freeze_one_row();
@@ -916,9 +945,9 @@ Ring::remove(row_t position)
  * Return: the newly added row.
  */
 VteRowData*
-Ring::append()
+Ring::append(guint8 bidi_flags)
 {
-	return insert(next());
+        return insert(next(), bidi_flags);
 }
 
 
@@ -1182,6 +1211,7 @@ Ring::rewrap(column_t columns,
 		/* Find the boundaries of the next paragraph */
 		gboolean prev_record_was_soft_wrapped = FALSE;
 		gboolean paragraph_is_ascii = TRUE;
+                guint8 paragraph_bidi_flags = old_record.bidi_flags;
 		gsize text_offset = paragraph_start_text_offset;
 		RowRecord new_record;
 		column_t col = 0;
@@ -1229,6 +1259,7 @@ Ring::rewrap(column_t columns,
 		new_record.text_start_offset = text_offset;
 		new_record.attr_start_offset = attr_offset;
 		new_record.is_ascii = paragraph_is_ascii;
+                new_record.bidi_flags = paragraph_bidi_flags;
 
 		while (paragraph_len > 0) {
 			/* Wrap one continuous run of identical attributes within the paragraph. */
